import random

from numpy import percentile, round, average


class Results(object):
    def __init__(self, results):
        self.results = results

    def percentile(self, pct):
        """What's the Xth percentile of the results.
        Arguments:
            pct(float): The percentile you'd like returned.
                Percentile to compute, which must be between 0 and 100 inclusive.
                Example: 50.0 == 50th percentile.
        Returns:
            int: The value from the results that that is at the percentile requested.
        """
        return percentile(self.results, pct, interpolation='nearest')


class Forecaster(object):
    def forecast(self, throughputs, backlog_size, num_simulations=10000, seed=None):
        """Forecasts how long a backlog will take to complete given the historical values provided.
        Arguments:
            throughputs(List[int]): Number of units completed per unit of time (stories per week, story points per month, etc.)
            backlog_size(int): Units in the backlog (stories, points, etc.)
        Returns:
            results
        Exceptions:
            None
        """
        results = []

        if seed is not None:
            random.seed(seed)

        for i in range(0, num_simulations):
            simulated_backlog = backlog_size
            time_unit_count = 0
            while simulated_backlog > 0:
                simulated_backlog -= random.choice(throughputs)
                time_unit_count += 1
            results.append(time_unit_count)

        return Results(results)

    def forecast_avg(self, throughputs, backlog_size):
        return backlog_size / average(throughputs)



import pytest


@pytest.fixture
def mut():
    """Returns the module under test."""
    from better import lib
    return lib


def test_results_percentile(mut):
    """Ensure the Results.percentile function returns the right value."""
    Results = mut.Results
    r = Results([
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10
    ])

    assert r.percentile(50.0) == 5


def test_forecast_average(mut):
    throughputs = [2, 2, 2, 2]
    backlog_size = 2 * 5

    result = mut.Forecaster().forecast_avg(throughputs, backlog_size)  # Fix the seed so we get predictable results
    assert result == 5


def test_forecast_sanity(mut):
    """With a uniform historical distribution, we should get the same result for all percentiles."""
    throughputs = [2, 2, 2, 2]
    backlog_size = 2 * 5

    result = mut.Forecaster().forecast(throughputs, backlog_size, seed=1)

    assert result.percentile(50.0) == 5


def test_forecast_with_backtest(mut):
    """With a known set of throughputs and a known backlog, we should get the result we expect."""
    throughputs = [
        14, 13, 1, 2, 12, 13, 7, 14
    ]
    backlog_size = sum(throughputs)

    result = mut.Forecaster().forecast(throughputs, backlog_size, seed=1)
    assert result.percentile(50.0) == 8
    assert result.percentile(75.0) == 10
    assert result.percentile(85.0) == 10
    assert result.percentile(95.0) == 12


def test_forecast_with_backtest_fewer_iters(mut):
    """With a known set of throughputs and a known backlog, we should get the result we expect."""
    throughputs = [
        14, 13, 1, 2, 12, 13, 7, 14
    ]
    backlog_size = sum(throughputs)

    result = mut.Forecaster().forecast(throughputs, backlog_size, num_simulations=10, seed=1)
    assert result.percentile(50.0) == 8
    assert result.percentile(75.0) == 9
    assert result.percentile(85.0) == 9
    assert result.percentile(95.0) == 11


def test_forecast_with_backtest_more_iters(mut):
    """With a known set of throughputs and a known backlog, we should get the result we expect."""
    throughputs = [
        14, 13, 1, 2, 12, 13, 7, 14
    ]
    backlog_size = sum(throughputs)

    result = mut.Forecaster().forecast(throughputs, backlog_size, num_simulations=50000, seed=1)
    assert result.percentile(50.0) == 8
    assert result.percentile(75.0) == 10
    assert result.percentile(85.0) == 10
    assert result.percentile(95.0) == 12



import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name = "better-project-forecast",
    version = "0.0.3",
    author = "Chris Heisel",
    author_email = "chris@heisel.org",
    description = ("Predict how long projects will take using monte carlo simulation"),
    license = "MIT",
    keywords = "monte carlo simulation project forecast",
    url = "https://github.com/cmheisel/better-project-forecast",
    packages=['better', 'tests'],
    long_description=read('README.md'),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
    ],
)
